@@ -5,6 +5,11 @@ Created on Sat Jul 20 12:07:57 2019
 @author: johnmount
 """
 
+from abc import ABC
+import math
+import pprint
+import warnings
+
 import numpy
 import pandas
 
@@ -12,9 +17,26 @@ import vtreat.util
 import vtreat.transform
 
 
-class VarTransform:
-    """build a treatment plan for a numeric outcome (regression)"""
+def ready_data_frame(d):
+    orig_type = type(d)
+    if orig_type == numpy.ndarray:
+        d = pandas.DataFrame(d)
+        d.columns = [str(c) for c in d.columns]
+    if not isinstance(d, pandas.DataFrame):
+        raise TypeError("not prepared to process type " + str(orig_type))
+    return d, orig_type
 
+
+def back_to_orig_type_data_frame(d, orig_type):
+    if not isinstance(d, pandas.DataFrame):
+        raise TypeError("Expected result to be a pandas.DataFrame, found: " + str(type(d)))
+    columns = [c for c in d.columns]
+    if orig_type == numpy.ndarray:
+        d = numpy.asarray(d)
+    return d, columns
+
+
+class VarTransform:
     def __init__(self, incoming_column_name, derived_column_names, treatment):
         self.incoming_column_name_ = incoming_column_name
         self.derived_column_names_ = derived_column_names.copy()
@@ -42,6 +64,7 @@ class MappedCodeTransform(VarTransform):
         res = pandas.merge(
             sf, self.code_book_, on=[self.incoming_column_name_], how="left", sort=False
         )  # ordered by left table rows
+        # could also try pandas .map()
         res = res[[derived_column_name]].copy()
         res.loc[vtreat.util.is_bad(res[derived_column_name]), derived_column_name] = 0
         return res
@@ -79,7 +102,7 @@ class CleanNumericTransform(VarTransform):
         self.replacement_value_ = replacement_value
 
     def transform(self, data_frame):
-        col = numpy.asarray(data_frame[self.incoming_column_name_].copy()).astype(float)
+        col = vtreat.util.safe_to_numeric_array(data_frame[self.incoming_column_name_])
         bad_posns = vtreat.util.is_bad(col)
         col[bad_posns] = self.replacement_value_
         res = pandas.DataFrame({self.derived_column_names_[0]: col})
@@ -98,6 +121,27 @@ class IndicateMissingTransform(VarTransform):
         return res.astype(float)
 
 
+def fit_clean_code(*, incoming_column_name, x, params, imputation_map):
+    if not vtreat.util.numeric_has_range(x):
+        return None
+    replacement = params['missingness_imputation']
+    try:
+        replacement = imputation_map[incoming_column_name]
+    except KeyError:
+        pass
+    if vtreat.util.can_convert_v_to_numeric(replacement):
+        replacement_value = 0.0 + replacement
+    elif callable(replacement):
+        replacement_value = vtreat.util.summarize_column(x, fn=replacement)
+    else:
+        raise TypeError("unexpected imputation type " + str(type(replacement)) + " (" + incoming_column_name + ")")
+    if pandas.isnull(replacement_value) or math.isnan(replacement_value) or math.isinf(replacement_value):
+        raise ValueError("replacement was bad " + incoming_column_name + ": " + str(replacement_value))
+    return CleanNumericTransform(
+            incoming_column_name=incoming_column_name, replacement_value=replacement_value
+        )
+
+
 def fit_regression_impact_code(*, incoming_column_name, x, y, extra_args, params):
     sf = vtreat.util.grouped_by_x_statistics(x, y)
     if sf.shape[0] <= 1:
@@ -142,7 +186,7 @@ def fit_regression_deviation_code(*, incoming_column_name, x, y, extra_args, par
 def fit_binomial_impact_code(*, incoming_column_name, x, y, extra_args, params):
     outcome_target = (extra_args["outcome_target"],)
     var_suffix = extra_args["var_suffix"]
-    y = numpy.asarray(numpy.asarray(y) == outcome_target, dtype=numpy.float64)
+    y = numpy.asarray(numpy.asarray(y) == outcome_target, dtype=float)
     sf = vtreat.util.grouped_by_x_statistics(x, y)
     if sf.shape[0] <= 1:
         return None
@@ -190,7 +234,7 @@ class IndicatorCodeTransform(VarTransform):
         def f(i):
             v = numpy.asarray(col == self.levels_[i]) + 0.0
             if self.sparse_indicators_:
-                v = pandas.SparseArray(v, fill_value=0.0)
+                v = pandas.arrays.SparseArray(v, fill_value=0.0)
             return v
 
         res = [
@@ -210,15 +254,16 @@ def fit_indicator_code(
     sf.loc[bad_posns, incoming_column_name] = "_NA_"
     counts = sf[incoming_column_name].value_counts()
     n = sf.shape[0]
+    counts = counts[counts > 0]
     counts = counts[counts >= min_fraction * n]  # no more than 1/min_fraction symbols
-    levels = [v for v in counts.index]
+    levels = [str(v) for v in counts.index]
     if len(levels) < 1:
         return None
     return IndicatorCodeTransform(
         incoming_column_name,
-        [incoming_column_name + "_lev_" + lev for lev in levels],
+        vtreat.util.build_level_codes(incoming_column_name, levels),
         levels=levels,
-        sparse_indicators=sparse_indicators,
+        sparse_indicators=sparse_indicators
     )
 
 
@@ -244,12 +289,14 @@ def fit_prevalence_code(incoming_column_name, x):
 
 # noinspection PyPep8Naming
 def fit_numeric_outcome_treatment(
-    *, X, y, var_list, outcome_name, cols_to_copy, params
+    *, X, y, var_list, outcome_name, cols_to_copy, params, imputation_map
 ):
     if (var_list is None) or (len(var_list) <= 0):
         var_list = [co for co in X.columns]
     copy_set = set(cols_to_copy)
     var_list = [co for co in var_list if (not (co in copy_set))]
+    v_counts = {v: vtreat.util.get_unique_value_count(X[v]) for v in var_list}
+    var_list = {v for v in var_list if v_counts[v] > 1}
     if len(var_list) <= 0:
         raise ValueError("no variables")
     xforms = []
@@ -269,15 +316,16 @@ def fit_numeric_outcome_treatment(
     var_list = [co for co in var_list if (not (co in set(all_bad)))]
     num_list = [co for co in var_list if vtreat.util.can_convert_v_to_numeric(X[co])]
     cat_list = [co for co in var_list if co not in set(num_list)]
+    id_like = [co for co in cat_list if v_counts[co] >= n]
+    if len(id_like) > 0:
+        warnings.warn("variable(s) " + ', '.join(id_like) + " have unique values per-row, dropping")
+        cat_list = [co for co in var_list if co not in set(id_like)]
     if "clean_copy" in params["coders"]:
         for vi in num_list:
-            summaryi = vtreat.util.characterize_numeric(X[vi])
-            if summaryi["varies"] and summaryi["has_range"]:
-                xforms = xforms + [
-                    CleanNumericTransform(
-                        incoming_column_name=vi, replacement_value=summaryi["mean"]
-                    )
-                ]
+            xform = fit_clean_code(incoming_column_name=vi, x=X[vi], params=params, imputation_map=imputation_map)
+            if xform is not None:
+                # noinspection PyTypeChecker
+                xforms = xforms + [xform]
     for vi in cat_list:
         if "impact_code" in params["coders"]:
             # noinspection PyTypeChecker
@@ -328,21 +376,23 @@ def fit_numeric_outcome_treatment(
 
 # noinspection PyPep8Naming
 def fit_binomial_outcome_treatment(
-    *, X, y, outcome_target, var_list, outcome_name, cols_to_copy, params
+    *, X, y, outcome_target, var_list, outcome_name, cols_to_copy, params, imputation_map
 ):
     if (var_list is None) or (len(var_list) <= 0):
         var_list = [co for co in X.columns]
     copy_set = set(cols_to_copy)
     var_list = [co for co in var_list if (not (co in copy_set))]
+    v_counts = {v: vtreat.util.get_unique_value_count(X[v]) for v in var_list}
+    var_list = {v for v in var_list if v_counts[v] > 1}
     if len(var_list) <= 0:
         raise ValueError("no variables")
     xforms = []
     n = X.shape[0]
-    all_badd = []
+    all_bad = []
     for vi in var_list:
         n_bad = sum(vtreat.util.is_bad(X[vi]))
         if n_bad >= n:
-            all_badd = all_badd + [vi]
+            all_bad = all_bad + [vi]
         if (n_bad > 0) and (n_bad < n):
             if "missing_indicator" in params["coders"]:
                 # noinspection PyTypeChecker
@@ -351,19 +401,19 @@ def fit_binomial_outcome_treatment(
                         incoming_column_name=vi, derived_column_name=vi + "_is_bad"
                     )
                 ]
-    var_list = [co for co in var_list if (not (co in set(all_badd)))]
+    var_list = [co for co in var_list if (not (co in set(all_bad)))]
     num_list = [co for co in var_list if vtreat.util.can_convert_v_to_numeric(X[co])]
     cat_list = [co for co in var_list if co not in set(num_list)]
+    id_like = [co for co in cat_list if v_counts[co] >= n]
+    if len(id_like) > 0:
+        warnings.warn("variable(s) " + ', '.join(id_like) + " have unique values per-row, dropping")
+        cat_list = [co for co in var_list if co not in set(id_like)]
     if "clean_copy" in params["coders"]:
         for vi in num_list:
-            summaryi = vtreat.util.characterize_numeric(X[vi])
-            if summaryi["varies"] and summaryi["has_range"]:
+            xform = fit_clean_code(incoming_column_name=vi, x=X[vi], params=params, imputation_map=imputation_map)
+            if xform is not None:
                 # noinspection PyTypeChecker
-                xforms = xforms + [
-                    CleanNumericTransform(
-                        incoming_column_name=vi, replacement_value=summaryi["mean"]
-                    )
-                ]
+                xforms = xforms + [xform]
     extra_args = {"outcome_target": outcome_target, "var_suffix": ""}
     for vi in cat_list:
         if "logit_code" in params["coders"]:
@@ -404,12 +454,14 @@ def fit_binomial_outcome_treatment(
 
 # noinspection PyPep8Naming
 def fit_multinomial_outcome_treatment(
-    *, X, y, var_list, outcome_name, cols_to_copy, params
+    *, X, y, var_list, outcome_name, cols_to_copy, params, imputation_map
 ):
     if (var_list is None) or (len(var_list) <= 0):
         var_list = [co for co in X.columns]
     copy_set = set(cols_to_copy)
     var_list = [co for co in var_list if (not (co in copy_set))]
+    v_counts = {v: vtreat.util.get_unique_value_count(X[v]) for v in var_list}
+    var_list = {v for v in var_list if v_counts[v] > 1}
     if len(var_list) <= 0:
         raise ValueError("no variables")
     xforms = []
@@ -431,16 +483,16 @@ def fit_multinomial_outcome_treatment(
     var_list = [co for co in var_list if (not (co in set(all_bad)))]
     num_list = [co for co in var_list if vtreat.util.can_convert_v_to_numeric(X[co])]
     cat_list = [co for co in var_list if co not in set(num_list)]
+    id_like = [co for co in cat_list if v_counts[co] >= n]
+    if len(id_like) > 0:
+        warnings.warn("variable(s) " + ', '.join(id_like) + " have unique values per-row, dropping")
+        cat_list = [co for co in var_list if co not in set(id_like)]
     if "clean_copy" in params["coders"]:
         for vi in num_list:
-            summaryi = vtreat.util.characterize_numeric(X[vi])
-            if summaryi["varies"] and summaryi["has_range"]:
+            xform = fit_clean_code(incoming_column_name=vi, x=X[vi], params=params, imputation_map=imputation_map)
+            if xform is not None:
                 # noinspection PyTypeChecker
-                xforms = xforms + [
-                    CleanNumericTransform(
-                        incoming_column_name=vi, replacement_value=summaryi["mean"]
-                    )
-                ]
+                xforms = xforms + [xform]
     for vi in cat_list:
         for outcome in outcomes:
             if "impact_code" in params["coders"]:
@@ -486,11 +538,13 @@ def fit_multinomial_outcome_treatment(
 
 
 # noinspection PyPep8Naming
-def fit_unsupervised_treatment(*, X, var_list, outcome_name, cols_to_copy, params):
+def fit_unsupervised_treatment(*, X, var_list, outcome_name, cols_to_copy, params, imputation_map):
     if (var_list is None) or (len(var_list) <= 0):
         var_list = [co for co in X.columns]
     copy_set = set(cols_to_copy)
     var_list = [co for co in var_list if (not (co in copy_set))]
+    v_counts = {v: vtreat.util.get_unique_value_count(X[v]) for v in var_list}
+    var_list = {v for v in var_list if v_counts[v] > 1}
     if len(var_list) <= 0:
         raise ValueError("no variables")
     xforms = []
@@ -511,16 +565,16 @@ def fit_unsupervised_treatment(*, X, var_list, outcome_name, cols_to_copy, param
     var_list = [co for co in var_list if (not (co in set(all_bad)))]
     num_list = [co for co in var_list if vtreat.util.can_convert_v_to_numeric(X[co])]
     cat_list = [co for co in var_list if co not in set(num_list)]
+    id_like = [co for co in cat_list if v_counts[co] >= n]
+    if len(id_like) > 0:
+        warnings.warn("variable(s) " + ', '.join(id_like) + " have unique values per-row, dropping")
+        cat_list = [co for co in var_list if co not in set(id_like)]
     if "clean_copy" in params["coders"]:
         for vi in num_list:
-            summaryi = vtreat.util.characterize_numeric(X[vi])
-            if summaryi["varies"] and summaryi["has_range"]:
+            xform = fit_clean_code(incoming_column_name=vi, x=X[vi], params=params, imputation_map=imputation_map)
+            if xform is not None:
                 # noinspection PyTypeChecker
-                xforms = xforms + [
-                    CleanNumericTransform(
-                        incoming_column_name=vi, replacement_value=summaryi["mean"]
-                    )
-                ]
+                xforms = xforms + [xform]
     for vi in cat_list:
         if "prevalence_code" in params["coders"]:
             # noinspection PyTypeChecker
@@ -528,6 +582,7 @@ def fit_unsupervised_treatment(*, X, var_list, outcome_name, cols_to_copy, param
                 fit_prevalence_code(incoming_column_name=vi, x=numpy.asarray(X[vi]))
             ]
         if "indicator_code" in params["coders"]:
+            # noinspection PyTypeChecker
             xforms = xforms + [
                 fit_indicator_code(
                     incoming_column_name=vi,
@@ -574,7 +629,7 @@ def pre_prep_frame(x, *, col_list, cols_to_copy):
             continue
         bad_ind = vtreat.util.is_bad(x[c])
         if vtreat.util.can_convert_v_to_numeric(x[c]):
-            x[c] = numpy.asarray(x[c] + 0, dtype=float)
+            x[c] = vtreat.util.safe_to_numeric_array(x[c])
         else:
             # https://stackoverflow.com/questions/22231592/pandas-change-data-type-of-series-to-string
             x[c] = numpy.asarray(x[c].apply(str), dtype=str)
@@ -584,11 +639,28 @@ def pre_prep_frame(x, *, col_list, cols_to_copy):
 
 def perform_transform(*, x, transform, params):
     plan = transform.plan_
-    new_frames = [xfi.transform(x) for xfi in plan["xforms"]]
-    for stp in params["user_transforms"]:
-        frm = stp.transform(X=x)
-        if frm is not None and frm.shape[1] > 0:
-            new_frames = new_frames + [frm]
+    xform_steps = [xfi for xfi in plan["xforms"]]
+    user_steps = [stp for stp in params["user_transforms"]]
+    # restrict down to to results we are going to use
+    if (transform.result_restriction is not None) and (len(transform.result_restriction) > 0):
+        xform_steps = [xfi for xfi in xform_steps
+                       if len(set(xfi.derived_column_names_).intersection(transform.result_restriction)) > 0]
+        user_steps = [stp for stp in user_steps
+                      if len(set(stp.derived_vars_).intersection(transform.result_restriction)) > 0]
+    # check all required columns are present
+    needs = set()
+    for xfi in xform_steps:
+        if xfi.incoming_column_name_ is not None:
+            needs.add(xfi.incoming_column_name_)
+    for stp in user_steps:
+        if stp.incoming_vars_ is not None:
+            needs.update(stp.incoming_vars_)
+    missing = needs - set(x.columns)
+    if len(missing) > 0:
+        raise ValueError("missing required input columns " + str(missing))
+    # do the work
+    new_frames = [xfi.transform(x) for xfi in (xform_steps + user_steps)]
+    new_frames = [frm for frm in new_frames if (frm is not None) and (frm.shape[1] > 0)]
     # see if we want to copy over any columns
     copy_set = set(plan["cols_to_copy"])
     to_copy = [ci for ci in x.columns if ci in copy_set]
@@ -605,30 +677,16 @@ def perform_transform(*, x, transform, params):
 def limit_to_appropriate_columns(*, res, transform):
     plan = transform.plan_
     to_copy = set(plan["cols_to_copy"])
-    if ("filter_to_recommended" in transform.params_.keys()) and transform.params_[
-        "filter_to_recommended"
-    ]:
-        to_take = set(
-            [
-                ci
-                for ci in transform.score_frame_["variable"][
-                    transform.score_frame_["recommended"]
-                ]
-            ]
-        )
-    else:
-        to_take = set(
-            [
-                ci
-                for ci in transform.score_frame_["variable"][
-                    transform.score_frame_["has_range"]
-                ]
-            ]
-        )
-    cols_to_keep = [ci for ci in res.columns if ci in to_copy or ci in to_take]
+    to_take = set([
+        ci for ci in transform.score_frame_["variable"][transform.score_frame_["has_range"]]])
+    if (transform.result_restriction is not None) and (len(transform.result_restriction) > 0):
+        to_take = to_take.intersection(transform.result_restriction)
+    cols_to_keep = [ci for ci in res.columns if (ci in to_copy) or (ci in to_take)]
     if len(cols_to_keep) <= 0:
         raise ValueError("no columns retained")
-    return res[cols_to_keep]
+    res = res[cols_to_keep].copy()
+    res.reset_index(inplace=True, drop=True)
+    return res
 
 
 # val_list is a list single column Pandas data frames
@@ -703,7 +761,7 @@ def cross_patch_refit_y_aware_cols(*, x, y, res, plan, cross_plan):
             cp = cross_plan[i]
             res.loc[cp["app"], derived_column_name] = numpy.asarray(
                 pi[derived_column_name]
-            ).reshape((len(pi)))
+            ).reshape((len(pi), ))
         res.loc[vtreat.util.is_bad(res[derived_column_name]), derived_column_name] = avg
     for xf in plan["xforms"]:
         xf.refitter_ = None
@@ -748,12 +806,14 @@ def cross_patch_user_y_aware_cols(*, x, y, res, params, cross_plan):
                     continue
                 pi.reset_index(inplace=True, drop=True)
                 cp = cross_plan[i]
-                res.loc[cp["app"], col] = numpy.asarray(pi[col]).reshape((len(pi)))
+                res.loc[cp["app"], col] = numpy.asarray(pi[col]).reshape((len(pi), ))
             res.loc[vtreat.util.is_bad(res[col]), col] = avg
     return res
 
 
-def score_plan_variables(cross_frame, outcome, plan, params):
+def score_plan_variables(cross_frame, outcome, plan, params,
+                         *,
+                         is_classification=False):
     def describe_xf(xf):
         description = pandas.DataFrame({"variable": xf.derived_column_names_})
         description["orig_variable"] = xf.incoming_column_name_
@@ -780,7 +840,10 @@ def score_plan_variables(cross_frame, outcome, plan, params):
     )
     var_table.reset_index(inplace=True, drop=True)
     sf = vtreat.util.score_variables(
-        cross_frame, variables=var_table["variable"], outcome=outcome
+        cross_frame,
+        variables=var_table["variable"],
+        outcome=outcome,
+        is_classification=is_classification
     )
     score_frame = pandas.merge(var_table, sf, how="left", on=["variable"], sort=False)
     num_treatment_types = len(score_frame["treatment"].unique())
@@ -838,12 +901,8 @@ def pseudo_score_plan_variables(*, cross_frame, plan, params):
     )
     score_frame.reset_index(inplace=True, drop=True)
 
-    def has_range(x):
-        x = numpy.asarray(x)
-        return numpy.max(x) > numpy.min(x)
-
     score_frame["has_range"] = [
-        has_range(cross_frame[c]) for c in score_frame["variable"]
+        vtreat.util.numeric_has_range(cross_frame[c]) for c in score_frame["variable"]
     ]
     score_frame["PearsonR"] = numpy.nan
     score_frame["significance"] = numpy.nan
@@ -852,3 +911,168 @@ def pseudo_score_plan_variables(*, cross_frame, plan, params):
     score_frame["vcount"] = score_frame.groupby("treatment")["_one"].transform("sum")
     score_frame.drop(["_one"], axis=1, inplace=True)
     return score_frame
+
+
+class VariableTreatment(ABC):
+    def __init__(
+            self, *,
+            var_list=None,
+            outcome_name=None,
+            outcome_target=None,
+            cols_to_copy=None,
+            params=None,
+            imputation_map=None,
+    ):
+        if var_list is None:
+            var_list = []
+        else:
+            var_list = vtreat.util.unique_itmes_in_order(var_list)
+        if cols_to_copy is None:
+            cols_to_copy = []
+        else:
+            cols_to_copy = vtreat.util.unique_itmes_in_order(cols_to_copy)
+        if outcome_name is not None and outcome_name not in set(cols_to_copy):
+            cols_to_copy = cols_to_copy + [outcome_name]
+        confused = set(cols_to_copy).intersection(set(var_list))
+        if len(confused) > 0:
+            raise ValueError("variables in treatment plan and non-treatment: " + ', '.join(confused))
+        if imputation_map is None:
+            imputation_map = {}  # dict
+        self.outcome_name_ = outcome_name
+        self.outcome_target_ = outcome_target
+        self.var_list_ = [vi for vi in var_list if vi not in set(cols_to_copy)]
+        self.cols_to_copy_ = cols_to_copy
+        self.params_ = params.copy()
+        self.imputation_map_ = imputation_map.copy()
+        self.plan_ = None
+        self.score_frame_ = None
+        self.cross_rows_ = None
+        self.cross_plan_ = None
+        self.last_fit_x_id_ = None
+        self.last_result_columns = None
+        self.result_restriction = None
+        self.clear()
+
+    def check_column_names(self, col_names):
+        to_check = set(self.var_list_)
+        if self.outcome_name_ is not None:
+            to_check.add(self.outcome_name_)
+        if self.cols_to_copy_ is not None:
+            to_check.update(self.cols_to_copy_)
+        seen = [c for c in col_names if c in to_check]
+        if len(seen) != len(set(seen)):
+            raise ValueError("duplicate column names in frame")
+
+    def clear(self):
+        self.plan_ = None
+        self.score_frame_ = None
+        self.cross_rows_ = None
+        self.cross_plan_ = None
+        self.last_fit_x_id_ = None
+        self.last_result_columns = None
+        self.result_restriction = None
+
+    def get_result_restriction(self):
+        if self.result_restriction is None:
+            return None
+        return self.result_restriction.copy()
+
+    def set_result_restriction(self, new_vars):
+        self.result_restriction = None
+        if (new_vars is not None) and (len(new_vars) > 0):
+            self.result_restriction = set(new_vars)
+
+    def merge_params(self, p):
+        raise NotImplementedError("base class called")
+
+    # display methods
+
+    def __repr__(self):
+        fmted = str(self.__class__.__module__) + "." + str(self.__class__.__name__) + '('
+        if self.outcome_name_ is not None:
+            fmted = fmted + "outcome_name=" + pprint.pformat(self.outcome_name_) + ", "
+        if self.outcome_target_ is not None:
+            fmted = fmted + "outcome_target=" + pprint.pformat(self.outcome_target_) + ", "
+        if (self.var_list_ is not None) and (len(self.var_list_) > 0):
+            fmted = fmted + "var_list=" + pprint.pformat(self.var_list_) + ", "
+        if (self.cols_to_copy_ is not None) and (len(self.cols_to_copy_) > 0):
+            fmted = fmted + "cols_to_copy=" + pprint.pformat(self.cols_to_copy_) + ", "
+        # if (self.params_ is not None) and (len(self.params_) > 0):
+        #     fmted = fmted + "params=" + pprint.pformat(self.params_) + ",\n"
+        # if (self.imputation_map_ is not None) and (len(self.imputation_map_) > 0):
+        #     fmted = fmted + "imputation_map=" + pprint.pformat(self.imputation_map_) + ",\n"
+        fmted = fmted + ')'
+        return fmted
+
+    def __str__(self):
+        return self.__repr__()
+
+    # sklearn pipeline step methods
+
+    # https://scikit-learn.org/stable/modules/generated/sklearn.base.TransformerMixin.html
+
+    # noinspection PyPep8Naming, PyUnusedLocal
+    def fit(self, X, y=None, **fit_params):
+        self.fit_transform(X=X, y=y)
+        return self
+
+    # noinspection PyPep8Naming, PyUnusedLocal
+    def fit_transform(self, X, y=None, **fit_params):
+        raise NotImplementedError("base class method called")
+
+    # noinspection PyPep8Naming
+    def transform(self, X):
+        raise NotImplementedError("base class method called")
+
+    # https://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html
+
+    # noinspection PyUnusedLocal,PyMethodMayBeStatic
+    def get_params(self, deep=True):
+        """
+        vtreat exposes a subset of controls as tunable parameters, users can choose this set
+        by specifying the tunable_params list in object construction parameters
+        """
+        return {ti: self.params_[ti] for ti in self.params_["tunable_params"]}
+
+    # noinspection PyUnusedLocal,PyMethodMayBeStatic
+    def set_params(self, **params):
+        """
+        vtreat exposes a subset of controls as tunable parameters, users can choose this set
+        by specifying the tunable_params list in object construction parameters
+        """
+        for (k, v) in params.items():
+            if k in self.params_["tunable_params"]:
+                self.params_[k] = v
+        return self
+
+    # extra methods to look more like sklearn objects
+
+    # https://scikit-learn.org/stable/modules/generated/sklearn.pipeline.Pipeline.html
+
+    # noinspection PyPep8Naming
+    def fit_predict(self, X, y=None, **fit_params):
+        return self.fit_transform(X=X, y=y, **fit_params)
+
+    # noinspection PyPep8Naming
+    def predict(self, X):
+        return self.transform(X)
+
+    # noinspection PyPep8Naming
+    def predict_proba(self, X):
+        return self.transform(X)
+
+    # https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/compose/_column_transformer.py
+
+    def get_feature_names(self, input_features=None):
+        if self.score_frame_ is None:
+            raise ValueError("get_feature_names called on uninitialized vtreat transform")
+        if self.params_['filter_to_recommended']:
+            new_vars = [self.score_frame_['variable'][i] for i in range(self.score_frame_.shape[0])
+                        if self.score_frame_['has_range'][i] and self.score_frame_['recommended'][i]
+                        and (input_features is None or self.score_frame_['orig_variable'][i] in input_features)]
+        else:
+            new_vars = [self.score_frame_['variable'][i] for i in range(self.score_frame_.shape[0])
+                        if self.score_frame_['has_range'][i]
+                        and (input_features is None or self.score_frame_['orig_variable'][i] in input_features)]
+        new_vars = new_vars + self.cols_to_copy_
+        return new_vars
